# -*- coding: utf-8 -*-
"""Pro Custodibus Agent.

Synchronizes your WireGuard settings with Pro Custodibus.

Usage:
  procustodibus-agent [--config=CONFIG] [--loop=SECONDS]
                      [-v | -vv | --verbosity=LEVEL]
  procustodibus-agent --help
  procustodibus-agent --test [--config=CONFIG] [-v | -vv | --verbosity=LEVEL]
  procustodibus-agent --version

Options:
  -h --help             Show this help
  --test                Run connectivity check
  --version             Show agent version
  -c --config=CONFIG    Config file
  -l --loop=SECONDS     Loop indefinitely, sending ping every SECONDS
  --verbosity=LEVEL     Log level (ERROR, WARNING, INFO, DEBUG)
  -v                    INFO verbosity
  -vv                   DEBUG verbosity
"""
import sys

from docopt import docopt

from procustodibus_agent import __version__ as version
from procustodibus_agent.agent import ping, ping_loop
from procustodibus_agent.cnf import Cnf
from procustodibus_agent.connectivity import check_connectivity


def main():
    """CLI Entry point."""
    args = docopt(__doc__)
    if args["--version"]:
        # print version to stdout
        print("procustodibus-agent " + version)  # noqa: T201
    elif args["--test"]:
        check(args["--config"], args["--verbosity"] or args["-v"])
    else:
        run(
            args["--config"],
            args["--verbosity"] or args["-v"],
            args["--loop"],
        )


def check(*args):
    """Runs connectivity check.

    Arguments:
        *args (list): List of arguments to pass to Cnf constructor.
    """
    cnf = Cnf(*args)
    sys.exit(check_connectivity(cnf))


def run(*args):
    """Runs CLI.

    Arguments:
        *args (list): List of arguments to pass to Cnf constructor.
    """
    cnf = Cnf(*args)

    if cnf.loop:
        ping_loop(cnf)
    else:
        ping(cnf)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""Routing and addressing utilities."""

from ipaddress import IPv4Address, IPv4Interface
from json import loads
from platform import system
from re import fullmatch, sub
from subprocess import PIPE, run  # noqa: S404


def run_ip_address_show(device):
    """Runs `ip address show` command for specified device.

    Arguments:
        device (str): Device name (eg 'wg0').

    Returns:
        dict: Parsed JSON from `ip --json address show` command.
    """
    args = ["ip", "--json", "address", "show", "dev", device]
    result = run(args, stdout=PIPE)  # noqa: S603 S607
    output = result.stdout.decode("utf-8")
    return loads(output)[0] if fullmatch(r"\[\{.*\}\]\s*", output) else {}


def run_ifconfig(device):
    """Runs `ifconfig` command for specified device.

    Arguments:
        device (str): Device name (eg 'wg0').

    Returns:
        list: Lines from `ifconfig` command.
    """
    args = ["ifconfig", device]
    result = run(args, stdout=PIPE)  # noqa: S603 S607
    output = result.stdout.decode("utf-8")
    return [x.strip() for x in output.split("\n")]


def annotate_wg_show_with_ip_address_show(interfaces):
    """Annotates parsed output of `wg show` with output of `ip address show`.

    Arguments:
        interfaces (dict): Dict parsed from `wg show` command.

    Returns:
        dict: Same dict with additional properties.
    """
    for name, properties in interfaces.items():
        _annotate_interface(name, properties)
    return interfaces


def _annotate_interface(name, properties):
    """Annotates specified dict with wg config for specified interface.

    Arguments:
        name (str): Interface name (eg 'wg0').
        properties (dict): Dict of interface properties.

    Returns:
        dict: Same dict with additional properties.
    """
    if system() == "Linux":
        _annotate_interface_with_ip_address_show(name, properties)
    else:
        _annotate_interface_with_ifconfig(name, properties)
    return properties


def _annotate_interface_with_ip_address_show(name, properties):
    """Annotates specified dict with wg config for specified interface.

    Arguments:
        name (str): Interface name (eg 'wg0').
        properties (dict): Dict of interface properties.

    Returns:
        dict: Same dict with additional properties.
    """
    info = run_ip_address_show(name)
    if info:
        properties["address"] = [_format_address_info(a) for a in info["addr_info"]]
    return properties


def _format_address_info(info):
    """Formats the addr_info object from `ip address show` as a CIDR.

    Arguments:
        info (dict): addr_info object.

    Returns:
        string: CIDR.
    """
    return "{}/{}".format(info["local"], info["prefixlen"])


def _annotate_interface_with_ifconfig(name, properties):
    """Annotates specified dict with wg config for specified interface.

    Arguments:
        name (str): Interface name (eg 'wg0').
        properties (dict): Dict of interface properties.

    Returns:
        dict: Same dict with additional properties.
    """
    addresses = list(filter(None, [_parse_inet_address(x) for x in run_ifconfig(name)]))
    if addresses:
        properties["address"] = addresses
    return properties


def _parse_inet_address(line):
    """Parses the inet or inet6 address out of the specified ifconfig output line.

    Arguments:
        line (str): ifconfig output line (eg 'inet6 ffc0:: prefixlen 64').

    Returns:
        str: CIDR or empty string (eg 'ffc0::/64').
    """
    if not line.startswith("inet"):
        return ""

    parts = line.split()
    if len(parts) < 4:
        return ""
    elif parts[2] == "-->" and len(parts) >= 6:
        del parts[2:4]

    return _format_inet_address_parts(parts)


def _format_inet_address_parts(parts):
    """Formats the specified ifconfig output tokens as a CIDR.

    Arguments:
        parts (list): ifconfig output tokens
            (eg ['inet6', 'ffc0::', 'prefixlen', '64']).

    Returns:
        str: CIDR or empty string (eg 'ffc0::/64').
    """
    if parts[1] == "addr:":
        return _strip_interface_id(parts[2])
    elif parts[1].startswith("addr:"):
        ip = parts[1][5:]
        mask = parts[3][5:] if parts[3].startswith("Mask:") else 32
        return str(IPv4Interface((ip, mask)))
    elif parts[2] == "prefixlen":
        ip = _strip_interface_id(parts[1])
        mask = parts[3]
        return f"{ip}/{mask}"
    elif parts[2] == "netmask":
        ip = _strip_interface_id(parts[1])
        mask = str(IPv4Address(int(parts[3], 16)))
        return str(IPv4Interface((ip, mask)))
    return ""


def _strip_interface_id(ip):
    """Strips the interface number or name from the specified IP address.

    Arguments:
        ip (str): IP address (eg 'fe80::169d:99ff:fe7f:8c67%utun0').

    Returns:
        str: IP address (eg 'fe80::169d:99ff:fe7f:8c67').
    """
    return sub(r"%[^/]*", "", ip)



# -*- coding: utf-8 -*-
"""Executor utilities."""

from logging import getLogger
from pathlib import Path
from re import match, search
from subprocess import PIPE, STDOUT, CompletedProcess, run  # noqa: S404
from tempfile import NamedTemporaryFile

from procustodibus_agent.resolve_hostname import split_endpoint_address_and_port
from procustodibus_agent.wg_cnf import (
    delete_wg_cnf,
    find_wg_cnf_path,
    rename_wg_cnf,
    update_wg_cnf,
)


def execute_desired(cnf, interfaces, data):
    """Executes desired changes from API.

    Arguments:
        cnf (Config): Config object.
        interfaces (dict): Interface info parsed from wg etc.
        data (list): List of desired change objects.

    Returns:
        list: Executed changes.
    """
    data = normalize_data(data)
    if not data:
        return []

    shut_down_interfaces(cnf, interfaces, data)
    delete_interfaces(cnf, interfaces, data)
    delete_endpoints(cnf, interfaces, data)
    update_interfaces(cnf, interfaces, data)
    update_routings(cnf, interfaces, data)
    update_endpoints(cnf, interfaces, data)
    create_interfaces(cnf, interfaces, data)
    create_routings(cnf, interfaces, data)
    create_endpoints(cnf, interfaces, data)
    rename_interfaces(cnf, interfaces, data)
    start_up_interfaces(cnf, interfaces, data)
    update_down_scripts(cnf, interfaces, data)

    return aggregate_executed_results(data)


def normalize_data(data):
    """Removes invalid desired change objects.

    And ensures they contain a fow core properties:
        type (str): Change type (eg 'desired_routings').
        id (str): Change ID (eg 'ABC123').
        attributes (dict): Properties to change.
        key (str): Log key to identify change (eg 'desired_interfaces ABC123').
        interface (str): Name of wg interface (eg 'wg0').
        results (list): List of change results (to append to).

    Arguments:
        data (list): List of desired change objects.

    Returns:
        list: Normalized list of desired change objects.
    """
    normalized = []

    for x in data:
        if (
            x.get("type")
            and x.get("id")
            and x.get("attributes")
            and (x["attributes"].get("name") or x["attributes"].get("interface"))
        ):
            x["key"] = f"{x['type']} {x['id']}"
            x["interface"] = x["attributes"].get("interface") or x["attributes"].get(
                "name"
            )
            x["results"] = []
            normalized.append(x)

    return normalized


def aggregate_executed_results(data):
    """Extracts result output from processed list of desired change objects.

    Arguments:
        data (list): Processed list of desired change objects.

    Returns:
        list: Flattened list of executed result output.
    """
    return [x for x in [executed_result(x) for x in data] if x]


def executed_result(change):
    """Extracts result output from a processed desired change object.

    Arguments:
        change (dict): Processed desired change object.

    Returns:
        list: Executed result output.
    """
    results = change["results"]
    output = []

    if next((x for x in results if x.returncode), None):
        output.append(f"fail {change['key']}")
    elif results:
        output.append(f"success {change['key']}")

    for x in results:
        if x.stdout:
            output.append(x.stdout)

    return "\n".join(output)


def shut_down_interfaces(cnf, interfaces, data):
    """Shuts down interfaces for desired change objects.

    Arguments:
        cnf (Config): Config object.
        interfaces (dict): Interface info parsed from wg etc.
        data (list): List of desired change objects.
    """
    for x in data:
        if _is_shut_down_interface(interfaces, x):
            shut_down_interface(cnf, interfaces, x)


def delete_interfaces(cnf, interfaces, data):
    """Deletes existing interfaces for desired change objects.

    Arguments:
        cnf (Config): Config object.
        interfaces (dict): Interface info parsed from wg etc.
        data (list): List of desired change objects.
    """
    for x in data:
        if _is_delete_interface(interfaces, x):
            delete_interface(cnf, interfaces, x)


def delete_endpoints(cnf, interfaces, data):
    """Deletes existing peers for desired change objects.

    Arguments:
        cnf (Config): Config object.
        interfaces (dict): Interface info parsed from wg etc.
        data (list): List of desired change objects.
    """
    for x in data:
        if _is_delete_endpoint(interfaces, x):
            delete_endpoint(cnf, interfaces, x)


def update_interfaces(cnf, interfaces, data):
    """Updates existing interfaces for desired change objects.

    Arguments:
        cnf (Config): Config object.
        interfaces (dict): Interface info parsed from wg etc.
        data (list): List of desired change objects.
    """
    for x in data:
        if _is_update_interface(interfaces, x):
            update_interface(cnf, interfaces, x)


def update_routings(cnf, interfaces, data):
    """Updates the routing info of existing interfaces for desired change objects.

    Arguments:
        cnf (Config): Config object.
        interfaces (dict): Interface info parsed from wg etc.
        data (list): List of desired change objects.
    """
    for x in data:
        if _is_update_routing(interfaces, x):
            update_routing(cnf, interfaces, x)


def update_endpoints(cnf, interfaces, data):
    """Updates existing peers for desired change objects.

    Arguments:
        cnf (Config): Config object.
        interfaces (dict): Interface info parsed from wg etc.
        data (list): List of desired change objects.
    """
    for x in data:
        if _is_update_endpoint(interfaces, x):
            update_endpoint(cnf, interfaces, x)


def create_interfaces(cnf, interfaces, data):
    """Creates new interfaces for desired change objects.

    Arguments:
        cnf (Config): Config object.
        interfaces (dict): Interface info parsed from wg etc.
        data (list): List of desired change objects.
    """
    for x in data:
        if _is_create_interface(interfaces, x):
            create_interface(cnf, interfaces, x)


def create_routings(cnf, interfaces, data):
    """Creates the routing info for new interfaces for desired change objects.

    Arguments:
        cnf (Config): Config object.
        interfaces (dict): Interface info parsed from wg etc.
        data (list): List of desired change objects.
    """
    for x in data:
        if _is_create_routing(interfaces, x):
            create_routing(cnf, interfaces, x)


def create_endpoints(cnf, interfaces, data):
    """Creates new peers for desired change objects.

    Arguments:
        cnf (Config): Config object.
        interfaces (dict): Interface info parsed from wg etc.
        data (list): List of desired change objects.
    """
    for x in data:
        if _is_create_endpoint(interfaces, x):
            create_endpoint(cnf, interfaces, x)


def rename_interfaces(cnf, interfaces, data):
    """Renames existing interfaces for desired change objects.

    Arguments:
        cnf (Config): Config object.
        interfaces (dict): Interface info parsed from wg etc.
        data (list): List of desired change objects.
    """
    for x in data:
        if _is_rename_interface(interfaces, x):
            rename_interface(cnf, interfaces, x)


# simpler for now to keep this logic together rather than subdivide it more functions
def start_up_interfaces(cnf, interfaces, data):  # noqa: CCR001
    """Starts up new or existing interfaces for desired change objects.

    Arguments:
        cnf (Config): Config object.
        interfaces (dict): Interface info parsed from wg etc.
        data (list): List of desired change objects.
    """
    desired_interfaces = {
        x["interface"]: x for x in data if x["type"] == "desired_interfaces"
    }

    to_start = {}
    for x in data:
        if _is_start_up_interface(interfaces, x):
            changes = to_start.get(x["interface"])
            if changes:
                changes.append(x)
            else:
                to_start[x["interface"]] = [x]

    for name, changes in to_start.items():
        desired = desired_interfaces.get(name)
        desired_up = desired and desired["attributes"].get("up")
        desired_down = desired and desired["attributes"].get("up") is False
        if not desired_down and (desired_up or interfaces.get(name)):
            start_up_interface(cnf, interfaces, name, changes)


def update_down_scripts(cnf, interfaces, data):
    """Updates the down scripts of existing interfaces for desired change objects.

    Arguments:
        cnf (Config): Config object.
        interfaces (dict): Interface info parsed from wg etc.
        data (list): List of desired change objects.
    """
    for x in data:
        if _is_update_down_script(interfaces, x):
            update_down_script(cnf, interfaces, x)


def shut_down_interface(cnf, interfaces, change):
    """Shuts down an existing interface from a desired change object.

    Arguments:
        cnf (Config): Config object.
        interfaces (dict): Interface info parsed from wg etc.
        change (dict): Desired change object.
    """
    results = change["results"]
    name = change["interface"]
    # remove to signal to other steps that the interface is down
    interface = interfaces.pop(name, {})
    manager = interface.get("manager") or cnf.manager

    # start back up at end if renaming and was previously up
    if (
        interface
        and change["attributes"].get("rename")
        and change["attributes"].get("up") is not False
    ):
        change["attributes"]["up"] = True

    if manager == "systemd":
        results.append(_cmd("systemctl", "stop", f"wg-quick@{name}.service"))
        results.append(_cmd("systemctl", "disable", f"wg-quick@{name}.service"))
    else:
        results.append(_cmd(cnf.wg_quick, "down", name))


def delete_interface(cnf, interfaces, change):
    """Deletes an existing interface from a desired change object.

    Arguments:
        cnf (Config): Config object.
        interfaces (dict): Interface info parsed from wg etc.
        change (dict): Desired change object.
    """
    results = change["results"]
    name = change["interface"]
    description = f"rm {find_wg_cnf_path(cnf, name)}"
    results.append(_as_cmd(description, lambda: delete_wg_cnf(cnf, name)))


def delete_endpoint(cnf, interfaces, change):
    """Deletes an existing peer from a desired change object.

    Arguments:
        cnf (Config): Config object.
        interfaces (dict): Interface info parsed from wg etc.
        change (dict): Desired change object.
    """
    results = change["results"]
    name = change["interface"]
    interface = interfaces.get(name)
    pk = change["attributes"].get("public_key")
    peer = interface["peers"].get(pk) if interface else None

    if peer:
        results.append(_cmd(cnf.wg, "set", name, "peer", pk, "remove"))
        _delete_endpoint_routes(interface, peer, change)

    if find_wg_cnf_path(cnf, name).exists():
        results.append(_update_wg_cnf_as_cmd(cnf, name, {}, [change["attributes"]]))
    else:
        results.append(CompletedProcess([], 0))


# simpler to keep this logic together rather than subdivide it more functions
def _delete_endpoint_routes(interface, peer, change):  # noqa: CCR001
    """Deletes an existing peer from a desired change object.

    Arguments:
        interface (dict): Interface info parsed from wg etc.
        peer (dict): Peer info parsed from wg etc.
        change (dict): Desired change object.
    """
    table = interface.get("table") or "auto"
    if table == "off":
        return

    allowed_ips = peer.get("allowed_ips") or []
    default_route = _uses_default_route(allowed_ips)

    if table != "auto" or not default_route:
        name = change["interface"]
        results = change["results"]
        for x in allowed_ips:
            _update_route("del", x, name, table, results)
    else:
        change["restart"] = True


def update_interface(cnf, interfaces, change):
    """Updates an existing interface from a desired change object.

    Arguments:
        cnf (Config): Config object.
        interfaces (dict): Interface info parsed from wg etc.
        change (dict): Desired change object.
    """
    to_set = {k: change["attributes"].get(k) for k in ["id", "description"]}
    results = change["results"]
    name = change["interface"]
    interface = interfaces.get(name)

    _update_interface_private_key(cnf, interface, change, name, to_set, results)

    port = change["attributes"].get("listen_port") or change["attributes"].get("port")
    if port is not None and port != interface.get("listen_port"):
        to_set["port"] = port
        results.append(_cmd(cnf.wg, "set", name, "listen-port", str(port)))

    fwmark = change["attributes"].get("fwmark")
    if fwmark is not None and fwmark != interface.get("fwmark"):
        to_set["fwmark"] = fwmark
        change["restart"] = True

    results.append(_update_wg_cnf_as_cmd(cnf, name, to_set))


def _update_interface_private_key(cnf, interface, change, name, to_set, results):
    """Updates the private key of an existing interface from a desired change object.

    Arguments:
        cnf (Config): Config object.
        interface (dict): Interface info parsed from wg etc.
        change (dict): Desired change object.
        name (str): Interface name (eg 'wg0').
        to_set (dict): Dict of interface properties to set (to add to).
        results (list): List of completed proccess objects (to add to).
    """
    pk = change["attributes"].get("private_key")
    if pk is not None and pk != interface.get("private_key"):
        to_set["private_key"] = pk
        if pk:
            with NamedTemporaryFile(mode="w", buffering=1) as f:
                f.write(pk)
                f.write("\n")
                results.append(_cmd(cnf.wg, "set", name, "private-key", f.name))
        else:
            results.append(_cmd(cnf.wg, "set", name, "private-key", "/dev/null"))


def update_routing(cnf, interfaces, change):
    """Updates the routing info for an existing interface from a desired change object.

    Arguments:
        cnf (Config): Config object.
        interfaces (dict): Interface info parsed from wg etc.
        change (dict): Desired change object.
    """
    to_set = {}
    results = change["results"]
    name = change["interface"]
    interface = interfaces.get(name)

    _update_routing_address(cnf, interface, change, name, to_set, results)
    _update_routing_dns(cnf, interface, change, name, to_set, results)
    _update_routing_mtu(cnf, interface, change, name, to_set, results)
    _update_routing_table(cnf, interface, change, name, to_set, results)
    _update_routing_scripts(cnf, interface, change, name, to_set, results)

    results.append(_update_wg_cnf_as_cmd(cnf, name, to_set))


# simpler to keep this logic together rather than subdivide it more functions
def _update_routing_address(  # noqa: CCR001
    cnf, interface, change, name, to_set, results
):
    """Updates the address list of an existing interface from a desired change object.

    Arguments:
        cnf (Config): Config object.
        interface (dict): Interface info parsed from wg etc.
        change (dict): Desired change object.
        name (str): Interface name (eg 'wg0').
        to_set (dict): Dict of interface properties to set (to add to).
        results (list): List of completed proccess objects (to add to).
    """
    new_addresses = change["attributes"].get("address")
    old_addresses = interface.get("address") or []
    if new_addresses is not None:
        for x in old_addresses:
            if x not in new_addresses:
                to_set["address"] = new_addresses
                family = calculate_ip_family(x)
                results.append(
                    _cmd("ip", f"-{family}", "address", "del", x, "dev", name)
                )
        for x in new_addresses:
            if x not in old_addresses:
                to_set["address"] = new_addresses
                family = calculate_ip_family(x)
                results.append(
                    _cmd("ip", f"-{family}", "address", "add", x, "dev", name)
                )


def _update_routing_dns(cnf, interface, change, name, to_set, results):
    """Updates the DNS list of an existing interface from a desired change object.

    Arguments:
        cnf (Config): Config object.
        interface (dict): Interface info parsed from wg etc.
        change (dict): Desired change object.
        name (str): Interface name (eg 'wg0').
        to_set (dict): Dict of interface properties to set (to add to).
        results (list): List of completed proccess objects (to add to).
    """
    dns = change["attributes"].get("dns")
    search = change["attributes"].get("search")
    if dns is None or search is None or (dns + search) == interface.get("dns"):
        return

    to_set["dns"] = dns
    to_set["search"] = search

    resolvconf_name = f"{resolvconf_interface_prefix()}{name}"
    results.append(_cmd("resolvconf", "-d", resolvconf_name, "-f"))
    resolvconf_input = []

    if dns:
        resolvconf_input.append(f"nameserver {','.join(dns)}\n")
    if search:
        resolvconf_input.append(f"search {','.join(search)}\n")
    if resolvconf_input:
        results.append(
            _cmd(
                "resolvconf",
                "-a",
                resolvconf_name,
                "-m",
                "0",
                "-x",
                stdin="".join(resolvconf_input),
            )
        )


def _update_routing_mtu(cnf, interface, change, name, to_set, results):
    """Updates the MTU of an existing interface from a desired change object.

    Arguments:
        cnf (Config): Config object.
        interface (dict): Interface info parsed from wg etc.
        change (dict): Desired change object.
        name (str): Interface name (eg 'wg0').
        to_set (dict): Dict of interface properties to set (to add to).
        results (list): List of completed proccess objects (to add to).
    """
    mtu = change["attributes"].get("mtu")
    if mtu is not None and mtu != interface.get("mtu"):
        to_set["mtu"] = mtu
        if mtu == 0:
            change["restart"] = True
        else:
            results.append(
                _cmd("ip", "link", "set", "mtu", str(mtu), "up", "dev", name)
            )


def _update_routing_table(cnf, interface, change, name, to_set, results):
    """Updates the routing table of an existing interface from a desired change object.

    Arguments:
        cnf (Config): Config object.
        interface (dict): Interface info parsed from wg etc.
        change (dict): Desired change object.
        name (str): Interface name (eg 'wg0').
        to_set (dict): Dict of interface properties to set (to add to).
        results (list): List of completed proccess objects (to add to).
    """
    table = change["attributes"].get("table")
    if table is not None and table != interface.get("table"):
        to_set["table"] = table
        # update_endpoint() needs to know the table to add routes to
        interface["new_table"] = table or "auto"
        change["restart"] = True


def _update_routing_scripts(cnf, interface, change, name, to_set, results):
    """Updates the scripts of an existing interface from a desired change object.

    Arguments:
        cnf (Config): Config object.
        interface (dict): Interface info parsed from wg etc.
        change (dict): Desired change object.
        name (str): Interface name (eg 'wg0').
        to_set (dict): Dict of interface properties to set (to add to).
        results (list): List of completed proccess objects (to add to).
    """
    for key in ["pre_up", "post_up", "save_config"]:
        value = change["attributes"].get(key)
        if value is not None and value != interface.get(key):
            to_set[key] = value
            if key != "save_config":
                change["restart"] = True


# TODO: refactor into smaller functions
def update_endpoint(cnf, interfaces, change):  # noqa: CCR001
    """Updates an existing peer from a desired change object.

    Arguments:
        cnf (Config): Config object.
        interfaces (dict): Interface info parsed from wg etc.
        change (dict): Desired change object.
    """
    to_set = {k: change["attributes"].get(k) for k in ["id", "name", "public_key"]}
    results = change["results"]
    name = change["interface"]
    interface = interfaces.get(name)
    pk = change["attributes"].get("public_key")
    peer = interface["peers"].get(pk) or {}

    old_table = interface.get("table") or "auto"
    new_table = interface.get("new_table") or old_table
    not_auto_table = old_table != "auto" and new_table != "auto"

    new_allowed_ips = change["attributes"].get("allowed_ips")
    old_allowed_ips = peer.get("allowed_ips") or []
    default_route = _uses_default_route(old_allowed_ips) or _uses_default_route(
        new_allowed_ips
    )
    allowed_ips_different = new_allowed_ips is not None and set(
        old_allowed_ips
    ).symmetric_difference(new_allowed_ips)

    if not_auto_table or not default_route or not allowed_ips_different:
        _update_endpoint_preshared_key(cnf, peer, change, name, pk, to_set, results)
        if allowed_ips_different:
            _update_endpoint_allowed_ips(
                cnf,
                old_allowed_ips,
                new_allowed_ips,
                name,
                pk,
                old_table,
                new_table,
                to_set,
                results,
            )
        _update_endpoint_address(cnf, peer, change, name, pk, to_set, results)

        keepalive = change["attributes"].get("keepalive")
        if keepalive is not None and keepalive != peer.get("persistent_keepalive"):
            to_set["keepalive"] = keepalive
            _wg_set_peer(cnf, name, pk, "persistent-keepalive", str(keepalive), results)

    else:
        to_set = change["attributes"]
        change["restart"] = True

    results.append(_update_wg_cnf_as_cmd(cnf, name, {}, [to_set]))


def _uses_default_route(ips):
    """True if the specified AllowedIPs list should override the default route.

    Arguments:
        ips (list): AllowedIPs list (eg ['0.0.0.0/0, ::/0']).

    Returns:
        boolean: True if should override the default route.
    """
    if ips:
        for ip in ips:
            if "/0" in ip:
                return True
    return False


# simpler to keep this logic together rather than subdivide it more functions
def _update_endpoint_preshared_key(  # noqa: CFQ002
    cnf, peer, change, name, pk, to_set, results
):
    """Updates the preshared key of an existing peer from a desired change object.

    Arguments:
        cnf (Config): Config object.
        peer (dict): Peer info parsed from wg etc.
        change (dict): Desired change object.
        name (str): Interface name (eg 'wg0').
        pk (str): Peer public key (eg 'ABC...123=').
        to_set (dict): Dict of peer properties to set (to add to).
        results (list): List of completed proccess objects (to add to).
    """
    shared = change["attributes"].get("preshared_key")
    if shared is not None and shared != peer.get("preshared_key"):
        if shared and shared != "off":
            to_set["preshared_key"] = shared
            with NamedTemporaryFile(mode="w", buffering=1) as f:
                f.write(shared)
                f.write("\n")
                _wg_set_peer(cnf, name, pk, "preshared-key", f.name, results)
        else:
            to_set["preshared_key"] = ""
            _wg_set_peer(cnf, name, pk, "preshared-key", "/dev/null", results)


# simpler to keep this logic together rather than subdivide it more functions
def _update_endpoint_allowed_ips(  # noqa: CCR001, CFQ002
    cnf,
    old_allowed_ips,
    new_allowed_ips,
    name,
    pk,
    old_table,
    new_table,
    to_set,
    results,
):
    """Updates the allowed IPs list an existing peer from a desired change object.

    Arguments:
        cnf (Config): Config object.
        old_allowed_ips (list): Old list of AllowedIPs.
        new_allowed_ips (list): New list of AllowedIPs.
        name (str): Interface name (eg 'wg0').
        pk (str): Peer public key (eg 'ABC...123=').
        old_table (str): Old route table name (eg 'auto').
        new_table (str): New route table name (eg 'off').
        to_set (dict): Dict of peer properties to set (to add to).
        results (list): List of completed proccess objects (to add to).
    """
    to_set["allowed_ips"] = new_allowed_ips
    _wg_set_peer(cnf, name, pk, "allowed-ips", ",".join(new_allowed_ips), results)

    if old_table != "off":
        for x in old_allowed_ips:
            if x not in new_allowed_ips:
                _update_route("del", x, name, old_table, results)

    if new_table != "off":
        for x in new_allowed_ips:
            if x not in old_allowed_ips:
                _update_route("add", x, name, new_table, results)


def _update_route(action, to, dev, table, results):
    """Adds or removes the specified route.

    Arguments:
        action (str): "add" or "del".
        to (str): IP address or CIDR (eg '10.0.0.0/24').
        dev (str): Interface name (eg 'wg0').
        table (str): Route table name (eg 'auto').
        results (list): List of completed proccess objects (to add to).
    """
    family = calculate_ip_family(to)

    args = ["ip", f"-{family}", "route", action, to, "dev", dev]
    if table != "auto":
        args += ["table", table]

    results.append(_cmd(*args))


# simpler to keep this logic together rather than subdivide it more functions
def _update_endpoint_address(  # noqa: CCR001, CFQ002
    cnf, peer, change, name, pk, to_set, results
):
    """Updates the endpoint address an existing peer from a desired change object.

    Arguments:
        cnf (Config): Config object.
        peer (dict): Peer info parsed from wg etc.
        change (dict): Desired change object.
        name (str): Interface name (eg 'wg0').
        pk (str): Peer public key (eg 'ABC...123=').
        to_set (dict): Dict of peer properties to set (to add to).
        results (list): List of completed proccess objects (to add to).
    """
    endpoint = change["attributes"].get("endpoint")
    if endpoint is None:
        return

    old_endpoint = peer.get("endpoint") or ""
    old_hostname = peer.get("hostname") or ""
    if old_hostname:
        old_ip, old_port = split_endpoint_address_and_port(old_endpoint)
        old_endpoint = (
            f"{old_hostname}:{old_port}" if old_port != 51820 else old_hostname
        )

    if endpoint != old_endpoint:
        to_set["endpoint"] = endpoint
        if endpoint:
            _wg_set_peer(cnf, name, pk, "endpoint", endpoint, results)


def _wg_set_peer(cnf, name, pk, setting, value, results):
    """Sets the specified wg setting for the specified peer.

    Arguments:
        cnf (Config): Config object.
        name (str): Interface name (eg 'wg0').
        pk (str): Peer public key.
        setting (str): Setting name (eg 'allowed-ips').
        value (str): Setting value (eg '10.0.0.1/24,fc00:0:0:1::/64').
        results (list): List to append the results of the command to.
    """
    results.append(_cmd(cnf.wg, "set", name, "peer", pk, setting, value))


def create_interface(cnf, interfaces, change):
    """Sets up a new interface from a desired change object.

    Arguments:
        cnf (Config): Config object.
        interfaces (dict): Interface info parsed from wg etc.
        change (dict): Desired change object.
    """
    change["restart"] = True
    results = change["results"]
    name = change["interface"]
    to_set = change["attributes"]
    results.append(_update_wg_cnf_as_cmd(cnf, name, to_set))


def create_routing(cnf, interfaces, change):
    """Sets up the routing info for a new interface from a desired change object.

    Arguments:
        cnf (Config): Config object.
        interfaces (dict): Interface info parsed from wg etc.
        change (dict): Desired change object.
    """
    create_interface(cnf, interfaces, change)


def create_endpoint(cnf, interfaces, change):
    """Sets up a new peer from a desired change object.

    Arguments:
        cnf (Config): Config object.
        interfaces (dict): Interface info parsed from wg etc.
        change (dict): Desired change object.
    """
    change["restart"] = True
    results = change["results"]
    name = change["interface"]
    to_set = change["attributes"]
    results.append(_update_wg_cnf_as_cmd(cnf, name, {}, [to_set]))


def rename_interface(cnf, interfaces, change):
    """Renames an existing interface from a desired change object.

    Arguments:
        cnf (Config): Config object.
        interfaces (dict): Interface info parsed from wg etc.
        change (dict): Desired change object.
    """
    results = change["results"]
    old_name = change["interface"]
    old_path = find_wg_cnf_path(cnf, old_name)
    new_name = change["attributes"].get("rename")

    if not new_name or new_name == old_name:
        update_interface(cnf, interfaces, change)
    elif old_path.exists():
        change["interface"] = new_name
        change["restart"] = True
        results.append(
            _as_cmd(
                f"mv {old_path} {find_wg_cnf_path(cnf, new_name)}",
                lambda: rename_wg_cnf(cnf, old_name, new_name),
            )
        )
    else:
        change["interface"] = new_name
        create_interface(cnf, interfaces, change)


def start_up_interface(cnf, interfaces, name, changes):
    """Starts up a new or existing interface for a list of desired changes.

    Arguments:
        cnf (Config): Config object.
        interfaces (dict): Interface info parsed from wg etc.
        name (str): Interface name to start up (eg `wg0`).
        changes (list): List of desired change objects.
    """
    results = []
    interface = interfaces.get(name) or {}
    manager = interface.get("manager") or cnf.manager

    if manager == "systemd":
        # stop first, and ignore shutdown output/errors
        _cmd("systemctl", "stop", f"wg-quick@{name}.service")
        results.append(_cmd("systemctl", "start", f"wg-quick@{name}.service"))
        results.append(_cmd("systemctl", "enable", f"wg-quick@{name}.service"))
    else:
        # stop first, and ignore shutdown output/errors
        _cmd(cnf.wg_quick, "down", name)
        results.append(_cmd(cnf.wg_quick, "up", name))

    for x in changes:
        x["results"].extend(results)


def update_down_script(cnf, interfaces, change):
    """Updates the routing info for an existing interface from a desired change object.

    Arguments:
        cnf (Config): Config object.
        interfaces (dict): Interface info parsed from wg etc.
        change (dict): Desired change object.
    """
    to_set = {}
    results = change["results"]
    name = change["interface"]
    interface = interfaces.get(name)

    for key in ["pre_down", "post_down"]:
        value = change["attributes"].get(key)
        if value is not None and value != interface.get(key):
            to_set[key] = value

    results.append(_update_wg_cnf_as_cmd(cnf, name, to_set))


def _cmd(*args, stdin=None):
    """Executes a command with the specified arguments.

    Arguments:
        *args (list): Command arguments.
        stdin (str): Optional command input.

    Returns:
        Completed process object.
    """
    try:
        if stdin:
            completed = run(  # noqa: S603 S607
                args,
                input=stdin,
                stdout=PIPE,
                stderr=STDOUT,
                timeout=30,
                universal_newlines=True,
            )
        else:
            completed = run(  # noqa: S603 S607
                args, stdout=PIPE, stderr=STDOUT, timeout=30, universal_newlines=True
            )
    except Exception as e:
        completed = CompletedProcess([], 1, stdout=str(e))

    output = " ".join(args)
    if completed.stdout:
        output = f"{output}\n{completed.stdout}"

    if completed.returncode:
        getLogger(__name__).warning("err # %s", output)
    else:
        getLogger(__name__).info("ok  # %s", output)

    return completed


def _as_cmd(description, function):
    """Runs the specified function and returns a CompletedProcess object as the result.

    Arguments:
        description (str): Command description (eg 'update config file').
        function (lambda): Zero-argument function to run.

    Returns:
        Completed process object.
    """
    try:
        function()
        getLogger(__name__).info("ok  # %s", description)
        return CompletedProcess([], 0)
    except Exception as e:
        getLogger(__name__).warning("err # %s", description, exc_info=True)
        return CompletedProcess([], 1, stdout=str(e))


def _update_wg_cnf_as_cmd(cnf, name, interface, peers=None):
    """Updates the cnf file for the specified interface with the specified properties.

    Arguments:
        cnf (Config): Config object.
        name (str): Name of interface to update (eg 'wg0').
        interface (dict): Properties of interface to update.
        peers (list): List of dicts with peer properties to update.

    Returns:
        Completed process object.
    """
    description = f"procustodibus_agent setconf {name} {find_wg_cnf_path(cnf, name)}"
    return _as_cmd(description, lambda: update_wg_cnf(cnf, name, interface, peers))


def _is_shut_down_interface(interfaces, change):
    """Checks if the specified change object should shut down an interface.

    Arguments:
        interfaces (dict): Interface info parsed from wg etc.
        change (dict): Desired change object.

    Returns:
        boolean: True if should shut down an interface.
    """
    return (
        change["type"] == "desired_interfaces"
        and interfaces.get(change["interface"])
        and (
            change["attributes"].get("up") is False
            or change["attributes"].get("delete")
            or change["attributes"].get("rename")
        )
    )


def _is_delete_interface(interfaces, change):
    """Checks if the specified change object should delete an interface.

    Arguments:
        interfaces (dict): Interface info parsed from wg etc.
        change (dict): Desired change object.

    Returns:
        boolean: True if should delete an interface
    """
    return change["type"] == "desired_interfaces" and change["attributes"].get("delete")


def _is_delete_endpoint(interfaces, change):
    """Checks if the specified change object should delete a peer.

    Arguments:
        interfaces (dict): Interface info parsed from wg etc.
        change (dict): Desired change object.

    Returns:
        boolean: True if should delete a peer.
    """
    return change["type"] == "desired_endpoints" and change["attributes"].get("delete")


def _is_update_interface(interfaces, change):
    """Checks if the specified change object should update an interface.

    Arguments:
        interfaces (dict): Interface info parsed from wg etc.
        change (dict): Desired change object.

    Returns:
        boolean: True if should update an interface.
    """
    return (
        change["type"] == "desired_interfaces"
        and not change["attributes"].get("delete")
        and interfaces.get(change["interface"])
    )


def _is_update_routing(interfaces, change):
    """Checks if the specified change object should update the routing for an interface.

    Arguments:
        interfaces (dict): Interface info parsed from wg etc.
        change (dict): Desired change object.

    Returns:
        boolean: True if should update the routing.
    """
    return change["type"] == "desired_routings" and interfaces.get(change["interface"])


def _is_update_endpoint(interfaces, change):
    """Checks if the specified change object should update a peer.

    Arguments:
        interfaces (dict): Interface info parsed from wg etc.
        change (dict): Desired change object.

    Returns:
        boolean: True if should update a peer.
    """
    return (
        change["type"] == "desired_endpoints"
        and not change["attributes"].get("delete")
        and interfaces.get(change["interface"])
    )


def _is_create_interface(interfaces, change):
    """Checks if the specified change object should create an interface.

    Arguments:
        interfaces (dict): Interface info parsed from wg etc.
        change (dict): Desired change object.

    Returns:
        boolean: True if should create an interface.
    """
    return (
        change["type"] == "desired_interfaces"
        and not change["attributes"].get("delete")
        and not change["attributes"].get("rename")
        and not interfaces.get(change["interface"])
    )


def _is_create_routing(interfaces, change):
    """Checks if the specified change object should create the routing for an interface.

    Arguments:
        interfaces (dict): Interface info parsed from wg etc.
        change (dict): Desired change object.

    Returns:
        boolean: True if should create the routing.
    """
    return change["type"] == "desired_routings" and not interfaces.get(
        change["interface"]
    )


def _is_create_endpoint(interfaces, change):
    """Checks if the specified change object should create a peer.

    Arguments:
        interfaces (dict): Interface info parsed from wg etc.
        change (dict): Desired change object.

    Returns:
        boolean: True if should create a peer.
    """
    return (
        change["type"] == "desired_endpoints"
        and not change["attributes"].get("delete")
        and not interfaces.get(change["interface"])
    )


def _is_rename_interface(interfaces, change):
    """Checks if the specified change object should rename an interface.

    Arguments:
        interfaces (dict): Interface info parsed from wg etc.
        change (dict): Desired change object.

    Returns:
        boolean: True if should rename an interface.
    """
    return (
        change["type"] == "desired_interfaces"
        and not change["attributes"].get("delete")
        and change["attributes"].get("rename")
    )


def _is_start_up_interface(interfaces, change):
    """Checks if the specified change object should start up an interface.

    Arguments:
        interfaces (dict): Interface info parsed from wg etc.
        change (dict): Desired change object.

    Returns:
        boolean: True if should start up an interface.
    """
    return (
        change.get("restart")
        or change["attributes"].get("restart")
        or (
            change["type"] == "desired_interfaces"
            and change["attributes"].get("up")
            and not change["attributes"].get("delete")
            and not interfaces.get(change["interface"])
        )
    )


def _is_update_down_script(interfaces, change):
    """Checks if the specified change object should update the down scripts.

    Arguments:
        interfaces (dict): Interface info parsed from wg etc.
        change (dict): Desired change object.

    Returns:
        boolean: True if should update the down scripts.
    """
    return change["type"] == "desired_routings" and interfaces.get(change["interface"])


def calculate_ip_family(address):
    """Determines whether the specified IP address is IPv4 or IPv6.

    Arguments:
        address (str): IP address (eg 'fc00:0:0:1::').

    Returns:
        int: 4 or 6.
    """
    return 6 if search(r":.*:", address) else 4


def resolvconf_interface_prefix(file="/etc/resolvconf/interface-order"):
    """Returns the prefix to append to an interface name for resolvconf commands.

    Arguments:
        file (str): Resolvconf interface-order file path to check.

    Returns:
        str: Prefix (eg 'tun') or blank ('').
    """
    path = Path(file)
    if path.exists():
        with open(path) as f:
            for line in f:
                interface_match = match(r"([A-Za-z0-9-]+)\*", line)
                if interface_match:
                    return interface_match[1]
    return ""



# -*- coding: utf-8 -*-
"""Utilities for manipulating wq conf ini."""

from contextlib import suppress
from pathlib import Path
from re import match, sub

from inflection import underscore

from procustodibus_agent import __version__ as version
from procustodibus_agent.api import format_datetime
from procustodibus_agent.cnf import (
    WIREGUARD_SPLITTABLE,
    find_ini_section_with_line,
    load_ini,
    load_ini_lines,
    replace_ini_line_value,
    save_ini_lines,
)


def annotate_wg_show_with_wg_cnf(cnf, interfaces):
    """Annotates parsed output of `wg show` with wg config.

    Arguments:
        cnf (Config): Config object.
        interfaces (dict): Dict parsed from `wg show` command.

    Returns:
        dict: Same dict with additional properties.
    """
    for name, properties in interfaces.items():
        _annotate_interface(cnf, name, properties)
    return interfaces


def find_wg_cnf_path(cnf, name):
    """Returns path to ini file for specified wg interface.

    Arguments:
        cnf (Config): Config object.
        name (str): Interface name (eg 'wg0').

    Returns:
        Path: Path to ini file (eg '/etc/wireguard/wg0.conf').
    """
    directory = cnf.wg_cnf_dir or "/etc/wireguard"
    return Path(directory, f"{name}.conf")


def _annotate_interface(cnf, name, properties):
    """Annotates specified dict with wg config for specified interface.

    Arguments:
        cnf (Config): Config object.
        name (str): Interface name (eg 'wg0').
        properties (dict): Dict of interface properties.

    Returns:
        dict: Same dict with additional properties.
    """
    path = find_wg_cnf_path(cnf, name)
    try:
        if not path.exists():
            return properties
    except PermissionError:
        return properties

    ini = load_ini(path, WIREGUARD_SPLITTABLE)
    interface = ini.get("interface")
    if not interface:
        return properties

    _annotate_interface_properties(properties, interface[0])
    _annotate_interface_peers(properties, ini)
    return properties


# keep conversion of all interface values together in same function
# even if it makes cognitive-complexity high
def _annotate_interface_properties(properties, ini):  # noqa: CCR001
    """Annotates specified interface properties from specified wg config.

    Arguments:
        properties (dict): Dict of interface properties.
        ini (dict): Dict of interface config from wg ini.

    Returns:
        dict: Same dict with additional properties.
    """
    # string properties
    for key in ["table"]:
        values = ini.get(key)
        if values:
            properties[key] = values[0]

    # integer properties
    for key in ["mtu"]:
        values = ini.get(key)
        if values:
            with suppress(ValueError):
                properties[key] = int(values[0])

    # boolean properties
    for key in ["save_config"]:
        values = ini.get(key)
        if values:
            properties[key] = values[0] == "true"

    # list properties
    for key in ["address", "dns", "pre_up", "post_up", "pre_down", "post_down"]:
        values = ini.get(key)
        if values:
            properties[key] = values

    return properties


def _annotate_interface_peers(interface, ini):
    """Annotates specified interface peers from specified wg config.

    Arguments:
        interface (dict): Dict of interface properties.
        ini (dict): Dict of full config from wg ini.

    Returns:
        dict: Same dict with additional properties.
    """
    peers = {p["public_key"][0]: p for p in ini.get("peer", []) if p.get("public_key")}
    for name, properties in interface.get("peers", {}).items():
        peer = peers.get(name)
        if peer:
            _annotate_peer_properties(properties, peer)
    return interface


def _annotate_peer_properties(properties, ini):
    """Annotates specified peer properties from specified wg config.

    Arguments:
        properties (dict): Dict of peer properties.
        ini (dict): Dict of peer config from wg ini.

    Returns:
        dict: Same dict with additional properties.
    """
    endpoint = ini.get("endpoint")
    if endpoint and match("[A-Za-z]", endpoint[0]):
        properties["hostname"] = sub(":.*", "", endpoint[0])

    return properties


def delete_wg_cnf(cnf, name):
    """Deletes the cnf file for the specified interface with the specified name.

    Arguments:
        cnf (Config): Config object.
        name (str): Interface name (eg 'wg0').
    """
    path = find_wg_cnf_path(cnf, name)
    if path.exists():
        path.unlink()


def rename_wg_cnf(cnf, old_name, new_name):
    """Renames the cnf file for the specified interface to the specified new name.

    Arguments:
        cnf (Config): Config object.
        old_name (str): Existing name (eg 'wg0').
        new_name (str): New name (eg 'wg100').
    """
    path = find_wg_cnf_path(cnf, old_name)
    if path.exists():
        path.rename(find_wg_cnf_path(cnf, new_name))


def update_wg_cnf(cnf, name, interface, peers=None):
    """Updates the cnf file for the specified interface with the specified properties.

    Arguments:
        cnf (Config): Config object.
        name (str): Name of interface to update (eg 'wg0').
        interface (dict): Properties of interface to update.
        peers (list): List of dicts with peer properties to update.

    Raises:
        ValueError: Cnf file for the interface cannot be written.
    """
    if not interface and not peers:
        return

    chmod = None
    path = find_wg_cnf_path(cnf, name)
    if path.exists():
        sections = load_ini_lines(path)
    elif interface.get("private_key"):
        sections = [_stub_wg_cnf_interface_section(cnf, interface)]
        chmod = 0o640
    else:
        raise ValueError(f"{path} does not exist and private key not available")

    section = _find_wg_cnf_interface_section(cnf, interface, path, sections)
    _update_wg_cnf_interface_section(section, interface)
    _update_wg_cnf_routing_section(section, interface)
    _update_wg_cnf_peer_sections(cnf, sections, peers)

    save_ini_lines(path, sections, chmod)


def _find_wg_cnf_interface_section(cnf, interface, path, sections):
    """Finds the interface section within the list of sections.

    Arguments:
        cnf (Config): Config object.
        interface (dict): Properties of interface to update.
        path (str): Path to cnf file.
        sections (list): List of list of lines.

    Returns:
        list: List of lines in interface section.

    Raises:
        ValueError: Interface section cannot be found or stubbed.
    """
    section = find_ini_section_with_line(sections, "[Interface]")
    if section:
        return section

    if interface.get("private_key"):
        section = _stub_wg_cnf_interface_section(cnf, interface)
        sections.append(section)
        return section
    else:
        raise ValueError(
            f"{path} does not contain an existing interface definition "
            "and private key not available"
        )


def _stub_wg_cnf_interface_section(cnf, interface):
    """Generates lines for the interface section of a new cnf file.

    Arguments:
        cnf (Config): Config object.
        interface (dict): Properties of interface.

    Returns:
        list: List of lines for new cnf file.
    """
    identifier = interface.get("id")
    description = interface.get("description")
    url = f"{cnf.app}/interfaces/{identifier}" if identifier else ""

    if description and url:
        description = f"{description} ({url})"
    else:
        description = description or url

    return [
        f"# generated {format_datetime()} by procustodibus-agent {version}",
        f"# {description}",
        "[Interface]",
        f"PrivateKey = {interface['private_key']}",
    ]


def _update_wg_cnf_interface_section(lines, interface):
    """Updates the lines of the interface section of a cnf file.

    Arguments:
        lines (list): List of lines in interface section.
        interface (dict): Core properties of interface to update.
    """
    pk = interface.get("private_key")
    if pk:
        replace_ini_line_value(lines, "PrivateKey", pk)

    _update_wg_cnf_number_line(lines, interface, "ListenPort", ["listen_port", "port"])
    _update_wg_cnf_number_line(lines, interface, "FwMark", ["fw_mark", "fwmark"])


def _update_wg_cnf_routing_section(lines, interface):
    """Updates the lines of the interface section of a cnf file with routing info.

    Arguments:
        lines (list): List of lines in interface section.
        interface (dict): Routing properties of interface to update.
    """
    _update_wg_cnf_list_line(lines, interface, "Address")

    dns = interface.get("dns")
    search = interface.get("search")
    if dns is not None and search is not None:
        dns = dns + search
    elif search is not None:
        dns = search
    if dns == []:
        replace_ini_line_value(lines, "DNS", None)
    elif dns:
        replace_ini_line_value(lines, "DNS", dns)

    _update_wg_cnf_number_line(lines, interface, "MTU")
    _update_wg_cnf_text_auto_line(lines, interface, "Table")

    for key in ["PreUp", "PostUp", "PreDown", "PostDown"]:
        _update_wg_cnf_list_line(lines, interface, key)

    _update_wg_cnf_boolean_line(lines, interface, "SaveConfig")


# keep top-level logic for updating peer sections together
# even if it makes cognitive-complexity high
def _update_wg_cnf_peer_sections(cnf, sections, peers):  # noqa: CCR001
    """Updates the lines of the peer sections of a cnf file.

    Arguments:
        cnf (Config): Config object.
        sections (list): List of list of lines.
        peers (list): List of dicts with peer properties to update.
    """
    for peer in peers or []:
        pk = peer["public_key"]
        section = find_ini_section_with_line(sections, f"PublicKey = {pk}")
        if peer.get("delete"):
            if section:
                sections.remove(section)
        else:
            if not section:
                section = _stub_wg_cnf_peer_section(cnf, peer)
                sections.append(section)
            _update_wg_cnf_peer_section(section, peer)


def _stub_wg_cnf_peer_section(cnf, peer):
    """Generates lines for a new peer section of a cnf file.

    Arguments:
        cnf (Config): Config object.
        peer (dict): Properties of the peer.

    Returns:
        list: List of lines for new peer.
    """
    identifier = peer.get("id")
    name = peer.get("name")
    url = f"{cnf.app}/endpoints/{identifier}" if identifier else ""

    if name and url:
        description = f"{name} ({url})"
    else:
        description = name or url

    return [
        "",
        f"# {description}",
        "[Peer]",
        f"PublicKey = {peer['public_key']}",
    ]


def _update_wg_cnf_peer_section(lines, peer):
    """Updates the lines of a peer section in a cnf file.

    Arguments:
        lines (list): Existing list of lines in the peer section.
        peer (dict): Properties of the peer to update.
    """
    _update_wg_cnf_text_off_line(lines, peer, "PresharedKey")
    _update_wg_cnf_list_line(lines, peer, "AllowedIPs", ["allowed_ips"])
    _update_wg_cnf_text_line(lines, peer, "Endpoint")
    _update_wg_cnf_number_line(
        lines, peer, "PersistentKeepalive", ["persistent_keepalive", "keepalive"]
    )


def _update_wg_cnf_list_line(lines, src, ini_key, src_keys=None):
    _update_wg_cnf_line(lines, src, [[]], ini_key, src_keys)


def _update_wg_cnf_text_line(lines, src, ini_key, src_keys=None):
    _update_wg_cnf_line(lines, src, [""], ini_key, src_keys)


def _update_wg_cnf_text_auto_line(lines, src, ini_key, src_keys=None):
    _update_wg_cnf_line(lines, src, ["", "auto"], ini_key, src_keys)


def _update_wg_cnf_text_off_line(lines, src, ini_key, src_keys=None):
    _update_wg_cnf_line(lines, src, ["", "off"], ini_key, src_keys)


def _update_wg_cnf_number_line(lines, src, ini_key, src_keys=None):
    _update_wg_cnf_line(lines, src, [0, "0", "off"], ini_key, src_keys)


def _update_wg_cnf_boolean_line(lines, src, ini_key, src_keys=None):
    _update_wg_cnf_line(lines, src, [False, "", "false"], ini_key, src_keys)


def _update_wg_cnf_line(lines, src, empty_values, ini_key, src_keys=None):
    """Updates a line with the specified key from the specified src dict.

    Arguments:
        lines (list): Existing list of lines.
        src (dict): Source dict from which to update the value.
        empty_values (list): Values that should result in removing the line
            instead of updating it.
        ini_key (str): Key in lines to update (eg 'ListenPort').
        src_keys (list): Src keys to check (eg ['listen_port', 'port']).
    """
    if not src_keys:
        src_keys = [underscore(ini_key)]

    value = None
    for x in src_keys:
        value = src.get(x)
        if value is not None:
            break

    if value in empty_values:
        replace_ini_line_value(lines, ini_key, None)
    elif value is True:
        replace_ini_line_value(lines, ini_key, "true")
    elif value:
        replace_ini_line_value(lines, ini_key, value)



# -*- coding: utf-8 -*-
"""MFA API utilities."""

import requests
from requests import Request

from procustodibus_agent.api import (
    API_TIMEOUT,
    encode_base64_web,
    raise_unless_has_cnf,
    send_with_session,
)
from procustodibus_agent.cnf import Cnf


def check_mfa_api(cnf, endpoint):
    """Checks if MFA for specified endpoint has expired.

    Arguments:
        cnf (Config): Config object.
        endpoint (str): Endpoint ID.

    Returns:
        str: OK, EXPIRED, UPDATING, or UNKNOWN.
    """
    if cnf is None:
        cnf = Cnf()
    response = requests.get(
        f"{cnf.api}/endpoints/{endpoint}/psk-rotation", timeout=API_TIMEOUT
    )
    response.raise_for_status()
    return response.text


def list_mfa_api(cnf):
    """Lists MFA state for all endpoints of the configured host.

    Arguments:
        cnf (Config): Config object.

    Returns:
        Response: Response json.
    """
    raise_unless_has_cnf(cnf)

    request = Request(
        "GET",
        f"{cnf.api}/endpoints",
        params={"included": "connection", "max": 1000},
    )

    response = send_with_session(cnf, request)
    return response.json()


def do_mfa_api(cnf, endpoint, user, token):
    """Synchorizes the MFA key of the specified endpoint.

    Arguments:
        cnf (Config): Config object.
        endpoint (str): Endpoint ID.
        user (str): User ID.
        token (str): Session token.

    Returns:
        Response: Response json.
    """
    url = f"{cnf.api}/endpoints/{endpoint}/psk-synchronize"
    authn = f"X-Custos user=^{user}, session=^{token}"

    response = requests.post(url, headers={"authorization": authn}, timeout=API_TIMEOUT)
    response.raise_for_status()
    return response.json()


def login_api_with_password(cnf, user, password, secondary_code=None):
    """Authenticates with specified username and password.

    Arguments:
        cnf (Config): Config object.
        user (str): User ID.
        password (str): Password.
        secondary_code (str): Optional secondary verification code.

    Returns:
        str: Session token.
    """
    raise_unless_has_cnf(cnf)

    url = f"{cnf.api}/sessions"
    password = encode_base64_web(password.encode("utf-8"))
    if secondary_code:
        code = encode_base64_web(secondary_code.encode("utf-8"))
        authn = f"X-Custos user=^{user}, password={password}, secondary_code={code}"
    else:
        authn = f"X-Custos user=^{user}, password={password}"

    response = requests.post(url, headers={"authorization": authn}, timeout=API_TIMEOUT)
    response.raise_for_status()
    return response.json()["data"][0]["attributes"]["token"]



# -*- coding: utf-8 -*-
"""Setuptools script for app."""
from setuptools import setup

from procustodibus_agent import __version__ as version

with open("README.md") as f:
    description = f.read()

with open("requirements/prod.txt") as f:
    requirements = f.read().splitlines()

with open("requirements/lint.txt") as f:
    requirements_lint = f.read().splitlines()

with open("requirements/test.txt") as f:
    requirements_test = f.read().splitlines()

setup(
    name="procustodibus_agent",
    version=version,
    description="Synchronizes your WireGuard settings with Pro Custodibus.",
    long_description=description,
    long_description_content_type="text/markdown",
    url="https://www.procustodibus.com/",
    project_urls={
        "Changelog": "https://docs.procustodibus.com/guide/agents/download/#changelog",
        "Documentation": "https://docs.procustodibus.com/guide/agents/run/",
        "Source": "https://git.sr.ht/~arx10/procustodibus-agent",
        "Tracker": "https://todo.sr.ht/~arx10/procustodibus-agent",
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: System Administrators",
        "Topic :: System :: Networking :: Monitoring",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    license="MIT",
    author="Arcem Tene",
    author_email="dev@arcemtene.com",
    packages=[
        "procustodibus_agent",
        "procustodibus_agent.mfa",
    ],
    entry_points={
        "console_scripts": [
            "procustodibus-agent = procustodibus_agent.cli:main",
            "procustodibus-credentials = procustodibus_agent.credentials:main",
            "procustodibus-mfa = procustodibus_agent.mfa.cli:main",
        ],
    },
    install_requires=requirements,
    extras_require={"lint": requirements_lint, "test": requirements_test},
    python_requires=">=3.6",
    zip_safe=True,
)



# -*- coding: utf-8 -*-
"""Pro Custodibus credentials tool.

Generates Pro Custodibus credentials from the configured agent-setup code.

Usage:
  procustodibus-credentials [--config=CONFIG] [-v | -vv | --verbosity=LEVEL]
  procustodibus-credentials --help
  procustodibus-credentials --version

Options:
  -h --help             Show this help
  --version             Show agent version
  -c --config=CONFIG    Config file
  --verbosity=LEVEL     Log level (ERROR, WARNING, INFO, DEBUG)
  -v                    INFO verbosity
  -vv                   DEBUG verbosity
"""
from docopt import docopt

from procustodibus_agent import __version__ as version
from procustodibus_agent.api import setup_api
from procustodibus_agent.cnf import Cnf


def main():
    """Tool entry point."""
    args = docopt(__doc__)
    if args["--version"]:
        # print version to stdout
        print("procustodibus-credentials " + version)  # noqa: T201
    else:
        run(
            args["--config"],
            args["--verbosity"] or args["-v"],
        )


def run(*args):
    """Runs tool.

    Arguments:
        *args (list): List of arguments to pass to Cnf constructor.
    """
    cnf = Cnf(*args)

    if not cnf.agent and type(cnf.setup) is dict:
        cnf.agent = cnf.setup.get("agent")

    if not cnf.host:
        cnf.host = "ignored"

    setup_api(cnf)


if __name__ == "__main__":
    main()
